"""
Implemented mass models
"""

import inspect

import numpy as np
import scipy.special as scs

from ..utils import powerlaw, trapezoid, truncnorm
from .interped import InterpolatedNoBaseModelIdentical

xp = np

__all__ = [
    "BaseSmoothedMassDistribution",
    "SinglePeakSmoothedMassDistribution",
    "MultiPeakSmoothedMassDistribution",
    "BrokenPowerLawSmoothedMassDistribution",
    "BrokenPowerLawPeakSmoothedMassDistribution",
    "InterpolatedPowerlaw",
    "double_power_law_primary_mass",
    "double_power_law_peak_primary_mass",
    "double_power_law_primary_power_law_mass_ratio",
    "power_law_primary_mass_ratio",
    "_primary_secondary_general",
    "power_law_primary_secondary_independent",
    "power_law_primary_secondary_identical",
    "power_law_mass",
    "two_component_single",
    "three_component_single",
    "two_component_primary_mass_ratio",
    "two_component_primary_secondary_independent",
    "two_component_primary_secondary_identical",
]


def double_power_law_primary_mass(mass, alpha_1, alpha_2, mmin, mmax, break_fraction):
    r"""
    Broken power-law mass distribution

    .. math::
        p(m | \alpha_1, m_\min, m_\max, \delta) &\propto \begin{cases}
            m^{-\alpha_1} : m_\min \leq m < m_\min + \delta (m_\max - m_\min)\\
            m^{-\alpha_2} : m_\min + \delta (m_\max - m_\min) \leq m < m_\max
        \end{cases}

    Parameters
    ----------
    mass: array-like
        Mass to evaluate probability at (:math:`m`).
    alpha_1: float
        Powerlaw exponent for more massive black hole below break (:math:`\alpha_1`).
    alpha_2: float
        Powerlaw exponent for more massive black hole above break (:math:`\alpha_2`).
    break_fraction: float
        The fraction between mmin and mmax primary mass distribution breaks (:math:`\delta`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum mass in the powerlaw distributed component (:math:`m_\max`).
    """

    m_break = mmin + break_fraction * (mmax - mmin)
    correction = powerlaw(m_break, alpha=-alpha_2, low=m_break, high=mmax) / powerlaw(
        m_break, alpha=-alpha_1, low=mmin, high=m_break
    )
    low_part = powerlaw(mass, alpha=-alpha_1, low=mmin, high=m_break)
    high_part = powerlaw(mass, alpha=-alpha_2, low=m_break, high=mmax)
    prob = low_part * (mass < m_break) * correction + high_part * (mass >= m_break)
    return prob / (1 + correction)


def double_power_law_peak_primary_mass(
    mass,
    alpha_1,
    alpha_2,
    mmin,
    mmax,
    break_fraction,
    lam,
    mpp,
    sigpp,
    gaussian_mass_maximum=100,
):
    r"""
    Broken power-law with a Gaussian component.

    .. math::
        p(m | \alpha_1, \alpha_2, m_\min, m_\max, \delta, \lambda_m, \mu_m, \sigma_m) =
        (1 - \lambda_m) p_{\text{bpl}}(m | \alpha_1, \alpha_2, m_\min, m_\max, \delta)
        + \lambda_m p_{\text{norm}}(m | \mu_m, \sigma_m)

    .. math::
        p_{\text{bpl}}(m | \alpha_1, m_\min, m_\max, \delta) &\propto \begin{cases}
            m^{-\alpha_1} : m_\min \leq m < m_\min + \delta (m_\max - m_\min)\\
            m^{-\alpha_2} : m_\min + \delta (m_\max - m_\min) \leq m < m_\max
        \end{cases}

    .. math::
        p_{\text{norm}}(m | \mu_m, \sigma_m) \propto \exp\left(-\frac{(m - \mu_{m})^2}{2\sigma^2_m}\right)

    Parameters
    ----------
    mass: array-like
        Mass to evaluate probability at (:math:`m`).
    alpha_1: float
        Powerlaw exponent for more massive black hole below break (:math:`\alpha_1`).
    alpha_2: float
        Powerlaw exponent for more massive black hole above break (:math:`\alpha_2`).
    break_fraction:float
        The fraction between mmin and mmax primary mass distribution breaks (:math:`\delta`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum mass in the powerlaw distributed component (:math:`m_\max`).
    lam: float
        Fraction of black holes in the Gaussian component (:math:`\lambda_m`).
    mpp: float
        Mean of the Gaussian component (:math:`\mu_m`).
    sigpp: float
        Standard deviation of the Gaussian component (:math:`\sigma_m`).
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
    """

    p_pow = double_power_law_primary_mass(
        mass=mass,
        alpha_1=alpha_1,
        alpha_2=alpha_2,
        mmin=mmin,
        mmax=mmax,
        break_fraction=break_fraction,
    )
    p_norm = truncnorm(mass, mu=mpp, sigma=sigpp, high=gaussian_mass_maximum, low=mmin)
    prob = (1 - lam) * p_pow + lam * p_norm
    return prob


def double_power_law_primary_power_law_mass_ratio(
    dataset, alpha_1, alpha_2, beta, mmin, mmax, break_fraction
):
    r"""
    Power law model for two-dimensional mass distribution, modelling primary
    mass and conditional mass ratio distribution.

    .. math::
        p(m_1, q) = p_{\text{bpl}}(m_1) p(q | m_1)

    .. math::
        p_{\text{bpl}}(m | \alpha_1, m_\min, m_\max, \delta) &\propto \begin{cases}
            m^{-\alpha_1} : m_\min \leq m < m_\min + \delta (m_\max - m_\min)\\
            m^{-\alpha_2} : m_\min + \delta (m_\max - m_\min) \leq m < m_\max
        \end{cases}

    .. math::
        p(q | m_1) \propto q^\beta : \frac{m_1}{m_\min} \leq q \leq 1

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for `mass_1` (:math:`m_1`) and `mass_ratio` (:math:`q`).
    alpha_1: float
        Negative power law exponent for more massive black hole before break (:math:`\alpha_1`).
    alpha_2: float
        Negative power law exponent for more massive black hole after break (:math:`\alpha_2`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    break_fraction: float
        Break point of the primary mass distribution.
        This is specified as a fraction of the way between mmin and mmax.
        E.g., mmin=5, mmax=45, break_fraction=0.5 would have a break at 25
    beta: float
        Power law exponent of the mass ratio distribution.
    """
    params = dict(mmin=mmin, mmax=mmax, break_fraction=break_fraction)
    p_m1 = double_power_law_primary_mass(
        dataset["mass_1"], alpha_1=alpha_1, alpha_2=alpha_2, **params
    )
    p_q = powerlaw(dataset["mass_ratio"], beta, 1, mmin / dataset["mass_1"])
    prob = p_m1 * p_q
    return prob


def power_law_primary_mass_ratio(dataset, alpha, beta, mmin, mmax):
    r"""
    Power law model for two-dimensional mass distribution, modelling primary
    mass and conditional mass ratio distribution.

    .. math::
        p(m_1, q) &= p_{\text{pow}}(m_1) p(q | m_1)

        p_{\text{pow}}(m) &\propto m^{-\alpha} : m_\min \leq m < m_\max

        p(q | m_1) &\propto q^\beta : \frac{m_1}{m_\min} \leq q \leq 1

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' (:math:`m_1`) and 'mass_ratio' (:math:`q`).
    alpha: float
        Negative power law exponent for more massive black hole (:math:`\alpha`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    beta: float
        Power law exponent of the mass ratio distribution (:math:`\beta`).
    """
    return two_component_primary_mass_ratio(
        dataset, alpha=alpha, beta=beta, mmin=mmin, mmax=mmax, lam=0, mpp=35, sigpp=1
    )


def _primary_secondary_general(dataset, p_m1, p_m2):
    return p_m1 * p_m2 * (dataset["mass_1"] >= dataset["mass_2"]) * 2


def power_law_primary_secondary_independent(dataset, alpha, beta, mmin, mmax):
    r"""
    Power law model for two-dimensional mass distribution, modelling the
    primary and secondary masses as following independent distributions.

    .. math::
        p(m1, m2) &= p_{\text{pow}}(m1) p_{\text{pow}}(m2) : m1 \geq m2

        p_{\text{pow}}(m) &\propto m^{-\alpha} : m_\min \leq m < m_\max

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' (:math:`m_1`) and 'mass_2' (:math:`m_2`).
    alpha: float
        Negative power law exponent for more massive black hole (:math:`\alpha`).
    beta: float
        Negative power law exponent of the secondary mass distribution (:math:`\beta`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    """
    p_m1 = powerlaw(dataset["mass_1"], -alpha, mmax, mmin)
    p_m2 = powerlaw(dataset["mass_2"], -beta, mmax, mmin)
    prob = _primary_secondary_general(dataset, p_m1, p_m2)
    return prob


def power_law_primary_secondary_identical(dataset, alpha, mmin, mmax):
    r"""
    Power law model for two-dimensional mass distribution, modelling the
    primary and secondary masses as following independent distributions.

    .. math::
        p(m_1, m_2 | \alpha, m_\min, m_\max) &= p_{\text{pow}}(m_1 | \alpha) p_{\text{pow}}(m_2 | \alpha) : m_1 \geq m_2

        p_{\text{pow}}(m | \alpha) &\propto m^{-\alpha} : m_\min \leq m < m_\max

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' (:math:`m_1`) and 'mass_2' (:math:`m_2`).
    alpha: float
        Negative power law exponent for both black holes (:math:`\alpha`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    """
    return power_law_primary_secondary_independent(
        dataset=dataset, alpha=alpha, beta=alpha, mmin=mmin, mmax=mmax
    )


def power_law_mass(mass, alpha, mmin, mmax):
    r"""
    Power law model for one-dimensional mass distribution.

    .. math::
        p(m) \propto m^{-\alpha} : m_\min \leq m < m_\max

    Parameters
    ----------
    mass: array-like
        Array of mass values (:math:`m`).
    alpha: float
        Negative power law exponent for the black hole distribution (:math:`\alpha`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    """
    return powerlaw(mass, alpha=-alpha, high=mmax, low=mmin)


def two_component_single(
    mass, alpha, mmin, mmax, lam, mpp, sigpp, gaussian_mass_maximum=100
):
    r"""
    Power law model for one-dimensional mass distribution with a Gaussian component.

    .. math::
        p(m) &= (1 - \lambda_m) p_{\text{pow}} + \lambda_m p_{\text{norm}}

        p_{\text{pow}}(m) &\propto m^{-\alpha} : m_\min \leq m < m_\max

        p_{\text{norm}}(m) &\propto \exp\left(-\frac{(m - \mu_{m})^2}{2\sigma^2_m}\right)

    Parameters
    ----------
    mass: array-like
        Array of mass values (:math:`m`).
    alpha: float
        Negative power law exponent for the black hole distribution (:math:`\alpha`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    lam: float
        Fraction of black holes in the Gaussian component (:math:`\lambda_m`).
    mpp: float
        Mean of the Gaussian component (:math:`\mu_m`).
    sigpp: float
        Standard deviation of the Gaussian component (:math:`\sigma_m`).
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
    """
    p_pow = powerlaw(mass, alpha=-alpha, high=mmax, low=mmin)
    p_norm = truncnorm(mass, mu=mpp, sigma=sigpp, high=gaussian_mass_maximum, low=mmin)
    prob = (1 - lam) * p_pow + lam * p_norm
    return prob


def three_component_single(
    mass,
    alpha,
    mmin,
    mmax,
    lam,
    lam_1,
    mpp_1,
    sigpp_1,
    mpp_2,
    sigpp_2,
    gaussian_mass_maximum=100,
):
    r"""
    Power law model for one-dimensional mass distribution with two Gaussian components.

    .. math::
        p(m) &= (1 - \lambda_m) p_{\text{pow}}(m) + \lambda_m \left(
            \hat{\lambda} p_{\text{norm}, 1}(m) + (1 - \hat{\lambda}) p_{\text{norm}, 2}(m)
        \right)

        p_{\text{pow}}(m) &\propto m^{-\alpha} : m_\min \leq m < m_\max

        p_{\text{norm}, i}(m) &\propto \exp\left(-\frac{(m - \mu_{m,i})^2}{2\sigma^2_{m, i}}\right)

    Parameters
    ----------
    mass: array-like
        Array of mass values (:math:`m`).
    alpha: float
        Negative power law exponent for the black hole distribution (:math:`\alpha`).
    mmin: float
        Minimum black hole mass (:math:`m_\min`).
    mmax: float
        Maximum black hole mass (:math:`m_\max`).
    lam: float
        Fraction of black holes in the Gaussian components (:math:`\lambda_m`).
    lam_1: float
        Fraction of black holes in the lower mass Gaussian component (:math:`\hat{\lambda}`).
    mpp_1: float
        Mean of the lower mass Gaussian component (:math:`\mu_{m, 1}`).
    mpp_2: float
        Mean of the upper mass Gaussian component (:math:`\mu_{m, 2}`).
    sigpp_1: float
        Standard deviation of the lower mass Gaussian component (:math:`\sigma_{m, 1}`).
    sigpp_2: float
        Standard deviation of the upper mass Gaussian component (:math:`\sigma_{m, 2}`).
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
        Note that this applies the same value to both.
    """
    p_pow = powerlaw(mass, alpha=-alpha, high=mmax, low=mmin)
    p_norm1 = truncnorm(
        mass, mu=mpp_1, sigma=sigpp_1, high=gaussian_mass_maximum, low=mmin
    )
    p_norm2 = truncnorm(
        mass, mu=mpp_2, sigma=sigpp_2, high=gaussian_mass_maximum, low=mmin
    )
    prob = (1 - lam) * p_pow + lam * lam_1 * p_norm1 + lam * (1 - lam_1) * p_norm2
    return prob


def two_component_primary_mass_ratio(
    dataset, alpha, beta, mmin, mmax, lam, mpp, sigpp, gaussian_mass_maximum=100
):
    r"""
    Power law model for two-dimensional mass distribution, modelling primary
    mass and conditional mass ratio distribution.

    .. math::
        p(m_1, q) = p(m1) p(q | m_1)

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha: float
        Negative power law exponent for more massive black hole.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum black hole mass.
    beta: float
        Power law exponent of the mass ratio distribution.
    lam: float
        Fraction of black holes in the Gaussian component.
    mpp: float
        Mean of the Gaussian component.
    sigpp: float
        Standard deviation of the Gaussian component.
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
    """
    params = dict(
        mmin=mmin,
        mmax=mmax,
        lam=lam,
        mpp=mpp,
        sigpp=sigpp,
        gaussian_mass_maximum=gaussian_mass_maximum,
    )
    p_m1 = two_component_single(dataset["mass_1"], alpha=alpha, **params)
    p_q = powerlaw(dataset["mass_ratio"], beta, 1, mmin / dataset["mass_1"])
    prob = p_m1 * p_q
    return prob


def two_component_primary_secondary_independent(
    dataset, alpha, beta, mmin, mmax, lam, mpp, sigpp, gaussian_mass_maximum=100
):
    r"""
    Power law model for two-dimensional mass distribution, modelling the
    primary and secondary masses as following independent distributions.

    .. math::
        p(m_1, m_2) = p_{\text{pow}}(m_1) p_{\text{pow}}(m_2) : m1 \geq m_2

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_2'.
    alpha: float
        Negative power law exponent for more massive black hole.
    beta: float
        Negative power law exponent of the secondary mass distribution.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum black hole mass.
    lam: float
        Fraction of black holes in the Gaussian component.
    mpp: float
        Mean of the Gaussian component.
    sigpp: float
        Standard deviation of the Gaussian component.
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
    """
    params = dict(
        mmin=mmin,
        mmax=mmax,
        lam=lam,
        mpp=mpp,
        sigpp=sigpp,
        gaussian_mass_maximum=gaussian_mass_maximum,
    )
    p_m1 = two_component_single(dataset["mass_1"], alpha=alpha, **params)
    p_m2 = two_component_single(dataset["mass_2"], alpha=beta, **params)

    prob = _primary_secondary_general(dataset, p_m1, p_m2)
    return prob


def two_component_primary_secondary_identical(
    dataset, alpha, mmin, mmax, lam, mpp, sigpp, gaussian_mass_maximum=100
):
    r"""
    Power law model for two-dimensional mass distribution, modelling the
    primary and secondary masses as following independent distributions.

    .. math::
        p(m_1, m_2) = p(m_1) p(m_2) : m_1 \geq m_2

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_2'.
    alpha: float
        Negative power law exponent for more massive black hole.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum black hole mass.
    lam: float
        Fraction of black holes in the Gaussian component.
    mpp: float
        Mean of the Gaussian component.
    sigpp: float
        Standard deviation of the Gaussian component.
    gaussian_mass_maximum: float, optional
        Upper truncation limit of the Gaussian component. (default: 100)
    """
    return two_component_primary_secondary_independent(
        dataset=dataset,
        alpha=alpha,
        beta=alpha,
        mmin=mmin,
        mmax=mmax,
        lam=lam,
        mpp=mpp,
        sigpp=sigpp,
        gaussian_mass_maximum=gaussian_mass_maximum,
    )


class BaseSmoothedMassDistribution:
    """
    Generic smoothed mass distribution base class.

    Implements the low-mass smoothing and power-law mass ratio
    distribution. Requires p_m1 to be implemented.

    Parameters
    ==========
    mmin: float
        The minimum mass considered for numerical normalization
    mmax: float
        The maximum mass considered for numerical normalization
    """

    primary_model = None

    @property
    def variable_names(self):
        vars = getattr(
            self.primary_model,
            "variable_names",
            inspect.getfullargspec(self.primary_model).args[1:],
        )
        vars += ["beta", "delta_m"]
        vars = set(vars).difference(self.kwargs.keys())
        return vars

    @property
    def kwargs(self):
        return dict()

    def __init__(self, mmin=2, mmax=100, normalization_shape=(1000, 500), cache=True):
        self.mmin = mmin
        self.mmax = mmax
        self.m1s = xp.linspace(mmin, mmax, normalization_shape[0])
        self.qs = xp.linspace(0.001, 1, normalization_shape[1])
        self.dm = self.m1s[1] - self.m1s[0]
        self.dq = self.qs[1] - self.qs[0]
        self.m1s_grid, self.qs_grid = xp.meshgrid(self.m1s, self.qs)
        self.cache = cache

    def __call__(self, dataset, *args, **kwargs):
        beta = kwargs.pop("beta")
        mmin = kwargs.get("mmin", self.mmin)
        mmax = kwargs.get("mmax", self.mmax)
        if "jax" not in xp.__name__:
            if mmin < self.mmin:
                raise ValueError(
                    "{self.__class__}: mmin ({mmin}) < self.mmin ({self.mmin})"
                )
            if mmax > self.mmax:
                raise ValueError(
                    "{self.__class__}: mmax ({mmax}) > self.mmax ({self.mmax})"
                )
        delta_m = kwargs.get("delta_m", 0)
        p_m1 = self.p_m1(dataset, **kwargs, **self.kwargs)
        p_q = self.p_q(dataset, beta=beta, mmin=mmin, delta_m=delta_m)
        prob = p_m1 * p_q
        return prob

    def p_m1(self, dataset, **kwargs):
        mmin = kwargs.get("mmin", self.mmin)
        delta_m = kwargs.pop("delta_m", 0)
        p_m = self.__class__.primary_model(dataset["mass_1"], **kwargs)
        p_m *= self.smoothing(
            dataset["mass_1"], mmin=mmin, mmax=self.mmax, delta_m=delta_m
        )
        norm = self.norm_p_m1(delta_m=delta_m, **kwargs)
        return p_m / norm

    def norm_p_m1(self, delta_m, **kwargs):
        """Calculate the normalisation factor for the primary mass"""
        mmin = kwargs.get("mmin", self.mmin)
        if "jax" not in xp.__name__ and delta_m == 0:
            return 1
        p_m = self.__class__.primary_model(self.m1s, **kwargs)
        p_m *= self.smoothing(self.m1s, mmin=mmin, mmax=self.mmax, delta_m=delta_m)

        norm = xp.nan_to_num(trapezoid(p_m, self.m1s)) * (delta_m != 0) + 1 * (
            delta_m == 0
        )
        return norm

    def p_q(self, dataset, beta, mmin, delta_m):
        p_q = powerlaw(dataset["mass_ratio"], beta, 1, mmin / dataset["mass_1"])
        p_q *= self.smoothing(
            dataset["mass_1"] * dataset["mass_ratio"],
            mmin=mmin,
            mmax=dataset["mass_1"],
            delta_m=delta_m,
        )

        try:
            if self.cache:
                p_q /= self.norm_p_q(beta=beta, mmin=mmin, delta_m=delta_m)
            else:
                self._cache_q_norms(dataset["mass_1"])
                p_q /= self.norm_p_q(beta=beta, mmin=mmin, delta_m=delta_m)
        except (AttributeError, TypeError, ValueError):
            self._cache_q_norms(dataset["mass_1"])
            p_q /= self.norm_p_q(beta=beta, mmin=mmin, delta_m=delta_m)

        return xp.nan_to_num(p_q)

    def norm_p_q(self, beta, mmin, delta_m):
        """Calculate the mass ratio normalisation by linear interpolation"""
        p_q = powerlaw(self.qs_grid, beta, 1, mmin / self.m1s_grid)
        p_q *= self.smoothing(
            self.m1s_grid * self.qs_grid, mmin=mmin, mmax=self.m1s_grid, delta_m=delta_m
        )

        norms = xp.nan_to_num(trapezoid(p_q, self.qs, axis=0)) * (delta_m != 0) + 1 * (
            delta_m == 0
        )

        return self._q_interpolant(norms)

    def _cache_q_norms(self, masses):
        """
        Cache the information necessary for linear interpolation of the mass
        ratio normalisation
        """
        from .interped import _setup_interpolant

        self._q_interpolant = _setup_interpolant(
            self.m1s, masses, kind="linear", backend=xp
        )

    @staticmethod
    def smoothing(masses, mmin, mmax, delta_m):
        """
        Apply a one sided window between mmin and mmin + delta_m to the
        mass pdf.

        The upper cut off is a step function,
        the lower cutoff is a logistic rise over delta_m solar masses.

        See T&T18 Eqs 7-8
        Note that there is a sign error in that paper.

        S = (f(m - mmin, delta_m) + 1)^{-1}
        f(m') = delta_m / m' + delta_m / (m' - delta_m)

        See also, https://en.wikipedia.org/wiki/Window_function#Planck-taper_window
        """
        if "jax" in xp.__name__ or delta_m > 0.0:
            shifted_mass = xp.nan_to_num((masses - mmin) / delta_m, nan=0)
            shifted_mass = xp.clip(shifted_mass, 1e-6, 1 - 1e-6)
            exponent = 1 / shifted_mass - 1 / (1 - shifted_mass)
            window = scs.expit(-exponent)
            window *= (masses >= mmin) * (masses <= mmax)
            return window
        else:
            return xp.ones(masses.shape)


class SinglePeakSmoothedMassDistribution(BaseSmoothedMassDistribution):
    """
    Powerlaw + peak model for two-dimensional mass distribution with low
    mass smoothing.

    https://arxiv.org/abs/1801.02699 Eq. (11) (T&T18)

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha: float
        Powerlaw exponent for more massive black hole.
    beta: float
        Power law exponent of the mass ratio distribution.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum mass in the powerlaw distributed component.
    lam: float
        Fraction of black holes in the Gaussian component.
    mpp: float
        Mean of the Gaussian component.
    sigpp: float
        Standard deviation of the Gaussian component.
    delta_m: float
        Rise length of the low end of the mass distribution.

    Notes
    -----
    The Gaussian component is bounded between [`mmin`, `self.mmax`].
    This means that the `mmax` parameter is _not_ the global maximum.
    """

    primary_model = two_component_single

    @property
    def kwargs(self):
        return dict(gaussian_mass_maximum=self.mmax)


class MultiPeakSmoothedMassDistribution(BaseSmoothedMassDistribution):
    """
    Powerlaw + two peak model for two-dimensional mass distribution with
    low mass smoothing.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha: float
        Powerlaw exponent for more massive black hole.
    beta: float
        Power law exponent of the mass ratio distribution.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum mass in the powerlaw distributed component.
    lam: float
        Fraction of black holes in the Gaussian component.
    lam_1: float
        Fraction of black holes in the lower mass Gaussian component.
    mpp_1: float
        Mean of the lower mass Gaussian component.
    mpp_2: float
        Mean of the upper mass Gaussian component.
    sigpp_1: float
        Standard deviation of the lower mass Gaussian component.
    sigpp_2: float
        Standard deviation of the upper mass Gaussian component.
    delta_m: float
        Rise length of the low end of the mass distribution.

    Notes
    -----
    The Gaussian components are bounded between [`mmin`, `self.mmax`].
    This means that the `mmax` parameter is _not_ the global maximum.
    """

    primary_model = three_component_single

    @property
    def kwargs(self):
        return dict(gaussian_mass_maximum=self.mmax)


class BrokenPowerLawSmoothedMassDistribution(BaseSmoothedMassDistribution):
    """
    Broken power law for two-dimensional mass distribution with low
    mass smoothing.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha_1: float
        Powerlaw exponent for more massive black hole below break.
    alpha_2: float
        Powerlaw exponent for more massive black hole above break.
    beta: float
        Power law exponent of the mass ratio distribution.
    break_fraction: float
        Fraction between mmin and mmax primary mass distribution breaks at.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum mass in the powerlaw distributed component.
    delta_m: float
        Rise length of the low end of the mass distribution.
    """

    primary_model = double_power_law_primary_mass


class BrokenPowerLawPeakSmoothedMassDistribution(BaseSmoothedMassDistribution):
    """
    Broken power law for two-dimensional mass distribution with low
    mass smoothing.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha_1: float
        Powerlaw exponent for more massive black hole below break.
    alpha_2: float
        Powerlaw exponent for more massive black hole above break.
    beta: float
        Power law exponent of the mass ratio distribution.
    break_fraction: float
        Fraction between mmin and mmax primary mass distribution breaks at.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum mass in the powerlaw distributed component.
    lam: float
        Fraction of black holes in the Gaussian component.
    mpp: float
        Mean of the Gaussian component.
    sigpp: float
        Standard deviation of the Gaussian component.
    delta_m: float
        Rise length of the low end of the mass distribution.

    Notes
    -----
    The Gaussian component is bounded between [`mmin`, `self.mmax`].
    This means that the `mmax` parameter is _not_ the global maximum.
    """

    primary_model = double_power_law_peak_primary_mass

    @property
    def kwargs(self):
        return dict(gaussian_mass_maximum=self.mmax)


class InterpolatedPowerlaw(
    BaseSmoothedMassDistribution, InterpolatedNoBaseModelIdentical
):
    """
    Interpolated powerlaw primary mass distribution with powerlaw mass ratio distribution.

    See https://arxiv.org/abs/2109.06137 for details.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'mass_1' and 'mass_ratio'.
    alpha: float
        Powerlaw exponent for more massive black hole.
    beta: float
        Power law exponent of the mass ratio distribution.
    mmin: float
        Minimum black hole mass.
    mmax: float
        Maximum mass in the powerlaw distributed component.
    delta_m: float
        Rise length of the low end of the mass distribution.
    mass{ii}: float
        The locations of the spline nodes for the primary mass distribution.
    fmass{ii}: float
        The values of the spline nodes for the primary mass distribution.
    """

    primary_model = power_law_mass

    def __init__(
        self,
        nodes=10,
        kind="cubic",
        mmin=2,
        mmax=100,
        normalization_shape=(1000, 500),
        regularize=False,
    ):
        """
        Parameters
        ==========
        nodes: int
            Number of spline nodes to use for interpolation, default=10.
        kind: str
            Order of the spline to use for interpolation, default="cubic".
        mmin: float
            The minimum mass considered for numerical normalization, default=2.
        mmax: float
            The maximum mass considered for numerical normalization, default=100.
        normalization_shape: tuple
            Shape of the grid used for numerical normalization, default=(1000, 500).
        regularize: bool
            Whether to regularize the spline node values to have root-mean-square value
            :code:`rms{name}`, default=False
        """
        BaseSmoothedMassDistribution.__init__(
            self,
            mmin=mmin,
            mmax=mmax,
            normalization_shape=normalization_shape,
        )
        InterpolatedNoBaseModelIdentical.__init__(
            self,
            minimum=mmin,
            maximum=mmax,
            parameters=["mass_1"],
            nodes=nodes,
            kind=kind,
            log_nodes=True,
            regularize=regularize,
        )
        self._xs = self.m1s

    @property
    def variable_names(self):
        variable_names = super().variable_names.union(
            InterpolatedNoBaseModelIdentical.variable_names.fget(self)
        )
        return variable_names

    def p_m1(self, dataset, **kwargs):

        f_splines, m_splines = self.extract_spline_points(kwargs)

        mmin = kwargs.get("mmin", self.mmin)
        delta_m = kwargs.pop("delta_m", 0)
        p_m = self.__class__.primary_model(
            dataset["mass_1"], **{key: kwargs[key] for key in ["alpha", "mmin", "mmax"]}
        )
        p_m *= self.smoothing(
            dataset["mass_1"], mmin=mmin, mmax=self.mmax, delta_m=delta_m
        )
        p_m *= self.p_x_unnormed(dataset, "mass_1", m_splines, f_splines, **kwargs)

        norm = self.norm_p_m1(delta_m=delta_m, f_splines=f_splines, **kwargs)
        return p_m / norm

    def norm_p_m1(self, delta_m, f_splines=None, **kwargs):
        mmin = kwargs.get("mmin", self.mmin)
        p_m = self.__class__.primary_model(
            self.m1s, **{key: kwargs[key] for key in ["alpha", "mmin", "mmax"]}
        )
        p_m *= self.smoothing(self.m1s, mmin=mmin, mmax=self.mmax, delta_m=delta_m) ** (
            delta_m > 0
        )
        p_m *= xp.exp(self._norm_spline(y=f_splines))
        norm = trapezoid(p_m, self.m1s)
        return norm



"""
Population models for :code:`GWPopulation` follow a specific syntax which is
an extension of the :code:`bilby.hyper.model` syntax.

Functions
---------

The simplest definition is as function that take as input a dictionary of
arrays with sample points named :code:`dataset` and named hyperparmeters
and return an array containing the probability density of the model
evaluated at the sample points.

For example, the following function defines a simple power-law mass model:

.. code-block:: python

    def power_law_mass_model(dataset, alpha, minimum_mass, maximum_mass):
        mass_1 = dataset["mass_1"]
        normalization = (
            (1 - alpha)
            / (maximum_mass ** (1 - alpha) - minimum_mass ** (1 - alpha))
        )
        in_bounds = (mass_1 >= minimum_mass) & (mass_1 <= maximum_mass)
        return mass_1 ** -alpha * normalization * in_bounds

Classes
-------

The more complex definition is as a class that has an :code:`__init__` method
that has the same signature. In this case a :code:`variable_names` attribute
or property can be used to avoid having to explicitly name the hyperparameters.
For example, the following class defines the same power-law mass model:

.. code-block:: python

    class PowerLawMassModel:

        variable_names = ["alpha", "minimum_mass", "maximum_mass"]

        def __init__(self, dataset, **kwargs):
            return power_law_mass_model(
                dataset,
                kwargs["alpha"],
                kwargs["minimum_mass"],
                kwargs["maximum_mass"],
            )

The class definition allows for more complex models to be defined and to cache
values that are expensive to compute and don't depend on the hyperparameters.

We provide a number of pre-defined models along with standard base models that
can be extended. See the API documentation for more details on implemented models.
"""

from . import mass, redshift, spin



import warnings
from copy import deepcopy
from functools import partial

import numpy as np
from bilby.core.likelihood import Likelihood


def generic_bilby_likelihood_function(likelihood, parameters, use_ratio=True):
    """
    A wrapper to allow a :code:`Bilby` likelihood to be used with :code:`jax`.

    Parameters
    ==========
    likelihood: bilby.core.likelihood.Likelihood
        The likelihood to evaluate.
    parameters: dict
        The parameters to evaluate the likelihood at.
    use_ratio: bool, optional
        Whether to evaluate the likelihood ratio or the full likelihood.
        Default is :code:`True`.
    """
    if use_ratio:
        return likelihood.log_likelihood_ratio(parameters)
    else:
        return likelihood.log_likelihood(parameters)


class JittedLikelihood(Likelihood):
    """
    A wrapper to just-in-time compile a :code:`Bilby` likelihood for use with :code:`jax`.

    .. note::

        This is currently hardcoded to return the log likelihood ratio, regardless of
        the input.

    Parameters
    ==========
    likelihood: bilby.core.likelihood.Likelihood
        The likelihood to wrap.
    likelihood_func: callable, optional
        The function to use to evaluate the likelihood. Default is
        :code:`generic_bilby_likelihood_function`. This function should take the
        likelihood and parameters as arguments along with additional keyword arguments.
    kwargs: dict, optional
        Additional keyword arguments to pass to the likelihood function.
    """

    def __init__(
        self, likelihood, likelihood_func=generic_bilby_likelihood_function, kwargs=None
    ):
        from jax import jit

        if kwargs is None:
            kwargs = dict()
        self.kwargs = kwargs
        self._likelihood = likelihood
        self.likelihood_func = jit(partial(likelihood_func, likelihood))
        super().__init__()

    def __getattr__(self, name):
        return getattr(self._likelihood, name)

    def log_likelihood_ratio(self, parameters):
        return float(np.nan_to_num(self.likelihood_func(parameters, **self.kwargs)))



import numpy as np
import pytest
from bilby.core.prior import DeltaFunction, Normal, PriorDict, Uniform

import gwpopulation
from gwpopulation.models import spin
from gwpopulation.utils import trapezoid, truncnorm

xp = np
N_TEST = 100


def tilt_prior():
    return PriorDict(dict(xi_spin=Uniform(0, 1), sigma_spin=Uniform(0, 4)))


def magnitude_prior():
    return PriorDict(
        dict(amax=Uniform(0.3, 1), alpha_chi=Uniform(1, 4), beta_chi=Uniform(1, 4))
    )


def tilt_test_data(xp):
    costilts = xp.linspace(-1, 1, 2000)
    dataset = dict(
        cos_tilt_1=xp.einsum("i,j->ij", costilts, xp.ones_like(costilts)),
        cos_tilt_2=xp.einsum("i,j->ji", costilts, xp.ones_like(costilts)),
    )
    return costilts, dataset


def magnitude_test_data(xp):
    a_array = xp.linspace(0, 1, 2000)
    dataset = dict(
        a_1=xp.einsum("i,j->ij", a_array, xp.ones_like(a_array)),
        a_2=xp.einsum("i,j->ji", a_array, xp.ones_like(a_array)),
    )
    return a_array, dataset


def test_spin_orientation_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    norms = list()
    prior = tilt_prior()
    costilts, dataset = tilt_test_data(xp)
    for ii in range(N_TEST):
        parameters = prior.sample()
        temp = spin.iid_spin_orientation_gaussian_isotropic(dataset, **parameters)
    norms.append(float(trapezoid(trapezoid(temp, costilts), costilts)))
    assert float(np.max(np.abs(1 - np.asarray(norms)))) < 1e-5


def test_iid_matches_independent_tilts(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    iid_params = dict(xi_spin=0.5, sigma_spin=0.5)
    ind_params = dict(xi_spin=0.5, sigma_1=0.5, sigma_2=0.5)
    _, dataset = tilt_test_data(xp)
    assert (
        xp.max(
            spin.iid_spin_orientation_gaussian_isotropic(dataset, **iid_params)
            - spin.independent_spin_orientation_gaussian_isotropic(
                dataset, **ind_params
            )
        )
        == 0
    )


def test_spin_magnitude_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    prior = magnitude_prior()
    a_array, dataset = magnitude_test_data(xp)
    norms = list()
    for ii in range(N_TEST):
        parameters = prior.sample()
        temp = spin.iid_spin_magnitude_beta(dataset, **parameters)
    norms.append(trapezoid(trapezoid(temp, a_array), a_array))
    assert float(xp.max(xp.abs(1 - xp.asarray(norms)))) < 1e-2


def test_returns_zero_alpha_beta_less_zero(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    prior = magnitude_prior()
    a_array, dataset = magnitude_test_data(xp)
    parameters = prior.sample()
    for key in ["alpha_chi", "beta_chi"]:
        parameters[key] = -1
        assert np.all(spin.iid_spin_magnitude_beta(dataset, **parameters) == 0)


def test_iid_matches_independent_magnitudes(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    prior = magnitude_prior()
    a_array, dataset = magnitude_test_data(xp)
    iid_params = prior.sample()
    ind_params = dict()
    ind_params.update({key + "_1": iid_params[key] for key in iid_params})
    ind_params.update({key + "_2": iid_params[key] for key in iid_params})
    assert (
        float(
            xp.max(
                spin.iid_spin_magnitude_beta(dataset, **iid_params)
                - spin.independent_spin_magnitude_beta(dataset, **ind_params)
            )
        )
        < 1e-5
    )


def test_iid_matches_independent(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    test_data = dict()
    costilts, new_data = tilt_test_data(xp)
    test_data.update(new_data)
    a_array, new_data = magnitude_test_data(xp)
    test_data.update(new_data)
    prior = magnitude_prior()
    prior.update(tilt_prior())
    params = prior.sample()
    mag_params = {key: params[key] for key in ["amax", "alpha_chi", "beta_chi"]}
    tilt_params = {key: params[key] for key in ["xi_spin", "sigma_spin"]}
    assert (
        float(
            xp.max(
                spin.iid_spin(test_data, **params)
                - spin.iid_spin_magnitude_beta(test_data, **mag_params)
                * spin.iid_spin_orientation_gaussian_isotropic(test_data, **tilt_params)
            )
        )
        < 1e-5
    )


def test_gaussian_chi_eff(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    assert (
        float(
            xp.max(
                spin.gaussian_chi_eff(
                    dict(chi_eff=xp.linspace(-2, 2, 1001)),
                    mu_chi_eff=0.4,
                    sigma_chi_eff=0.1,
                )
                - truncnorm(xp.linspace(-2, 2, 1001), mu=0.4, sigma=0.1, low=-1, high=1)
            )
        )
        < 1e-5
    )


def test_gaussian_chi_p(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    assert (
        float(
            xp.max(
                spin.gaussian_chi_p(
                    dict(chi_p=xp.linspace(-2, 2, 1001)), mu_chi_p=0.4, sigma_chi_p=0.1
                )
                - truncnorm(xp.linspace(-2, 2, 1001), mu=0.4, sigma=0.1, low=0, high=1)
            )
        )
        == 0
    )


def test_2d_gaussian_normalized(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = spin.GaussianChiEffChiP()
    chi_eff, chi_p = xp.meshgrid(xp.linspace(-1, 1, 501), xp.linspace(0, 1, 300))
    parameters = dict(
        mu_chi_eff=0.1,
        mu_chi_p=0.3,
        sigma_chi_eff=0.6,
        sigma_chi_p=0.5,
        spin_covariance=0.9,
    )
    prob = model(dict(chi_eff=chi_eff, chi_p=chi_p), **parameters)
    assert (
        xp.max(
            xp.abs(
                trapezoid(
                    trapezoid(prob, xp.linspace(-1, 1, 501)), xp.linspace(0, 1, 300)
                )
                - 1
            )
        )
        < 1e-3
    )


def test_2d_gaussian_no_covariance_matches_independent(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = spin.GaussianChiEffChiP()
    chi_eff, chi_p = xp.meshgrid(xp.linspace(-1, 1, 501), xp.linspace(0, 1, 300))
    data = dict(chi_eff=chi_eff, chi_p=chi_p)
    assert (
        xp.max(
            xp.abs(
                spin.gaussian_chi_eff(data, mu_chi_eff=0.6, sigma_chi_eff=0.2)
                * spin.gaussian_chi_p(data, mu_chi_p=0.4, sigma_chi_p=0.1)
                - model(
                    data,
                    mu_chi_eff=0.6,
                    mu_chi_p=0.4,
                    sigma_chi_eff=0.2,
                    sigma_chi_p=0.1,
                    spin_covariance=0.0,
                )
            )
        )
        < 1e-3
    )


# Helper functions for spline model tests


def spline_magnitude_prior(n_nodes, minimum=0, maximum=1):
    """Create prior for SplineSpinMagnitudeIdentical with n_nodes."""
    prior = PriorDict()
    nodes = np.linspace(minimum, maximum, n_nodes)
    for ii in range(n_nodes):
        prior[f"a{ii}"] = DeltaFunction(nodes[ii])
        prior[f"fa{ii}"] = Normal(0, 1)
    return prior


def spline_tilt_prior(n_nodes, minimum=-1, maximum=1):
    """Create prior for SplineSpinTiltIdentical with n_nodes."""
    prior = PriorDict()
    nodes = np.linspace(minimum, maximum, n_nodes)
    for ii in range(n_nodes):
        prior[f"cos_tilt{ii}"] = DeltaFunction(nodes[ii])
        prior[f"fcos_tilt{ii}"] = Normal(0, 1)
    return prior


# Tests for SplineSpinMagnitudeIdentical


@pytest.mark.parametrize("nodes", [4, 5, 7])
def test_spline_spin_magnitude_normalised(backend, nodes):
    """Test that SplineSpinMagnitudeIdentical is properly normalized."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = spin.SplineSpinMagnitudeIdentical(nodes=nodes)
    prior = spline_magnitude_prior(nodes)
    a_array, dataset = magnitude_test_data(xp)

    norms = list()
    # Use fewer iterations for computational efficiency
    n_test = 10
    for _ in range(n_test):
        parameters = prior.sample()
        prob = model(dataset, **parameters)
        norm = float(trapezoid(trapezoid(prob, a_array), a_array))
        norms.append(norm)

    # Allow 5% error tolerance for numerical integration
    assert xp.max(xp.abs(1 - xp.asarray(norms))) < 0.05


@pytest.mark.parametrize("kind", ["linear", "cubic"])
def test_spline_spin_magnitude_interpolation_kind(backend, kind):
    """Test that different interpolation kinds work for SplineSpinMagnitudeIdentical."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinMagnitudeIdentical(nodes=nodes, kind=kind)
    prior = spline_magnitude_prior(nodes)
    a_array, dataset = magnitude_test_data(xp)

    parameters = prior.sample()
    prob = model(dataset, **parameters)
    norm = float(trapezoid(trapezoid(prob, a_array), a_array))

    # Should be normalized regardless of interpolation kind
    # Allow 5% error tolerance for numerical integration
    assert abs(norm - 1.0) < 0.05


@pytest.mark.parametrize("regularize", [True, False])
def test_spline_spin_magnitude_regularization(backend, regularize):
    """Test that regularization option works for SplineSpinMagnitudeIdentical."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinMagnitudeIdentical(nodes=nodes, regularize=regularize)
    prior = spline_magnitude_prior(nodes)

    if regularize:
        prior["rmsa"] = Uniform(0.1, 2.0)

    a_array, dataset = magnitude_test_data(xp)
    parameters = prior.sample()

    # Should not raise an error
    prob = model(dataset, **parameters)

    # Check that variable names match expected
    if regularize:
        assert "rmsa" in model.variable_names
    else:
        assert "rmsa" not in model.variable_names


def test_spline_spin_magnitude_zero_outside_range(backend):
    """Test that SplineSpinMagnitudeIdentical returns zero outside the valid range."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    minimum, maximum = 0.2, 0.8
    model = spin.SplineSpinMagnitudeIdentical(
        minimum=minimum, maximum=maximum, nodes=nodes
    )
    prior = spline_magnitude_prior(nodes, minimum=minimum, maximum=maximum)

    # Create test data that includes values outside the range
    a_array = xp.linspace(0, 1, 2000)
    dataset = dict(
        a_1=xp.einsum("i,j->ij", a_array, xp.ones_like(a_array)),
        a_2=xp.einsum("i,j->ji", a_array, xp.ones_like(a_array)),
    )

    parameters = prior.sample()
    prob = model(dataset, **parameters)

    # Values outside [0.2, 0.8] should be zero
    prob_np = gwpopulation.utils.to_numpy(prob)
    a_array_np = gwpopulation.utils.to_numpy(a_array)

    assert np.max(prob_np[(a_array_np < minimum) | (a_array_np > maximum)]) == 0.0


def test_spline_spin_magnitude_variable_names():
    """Test that variable_names property works for SplineSpinMagnitudeIdentical."""
    nodes = 5
    model = spin.SplineSpinMagnitudeIdentical(nodes=nodes, regularize=False)

    # Should have node positions and values
    expected = [f"a{ii}" for ii in range(nodes)] + [f"fa{ii}" for ii in range(nodes)]
    assert set(model.variable_names) == set(expected)

    # With regularization, should also have rms parameter
    model_reg = spin.SplineSpinMagnitudeIdentical(nodes=nodes, regularize=True)
    expected_reg = expected + ["rmsa"]
    assert set(model_reg.variable_names) == set(expected_reg)


# Tests for SplineSpinTiltIdentical


@pytest.mark.parametrize("nodes", [4, 5, 7])
def test_spline_spin_tilt_normalised(backend, nodes):
    """Test that SplineSpinTiltIdentical is properly normalized."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = spin.SplineSpinTiltIdentical(nodes=nodes)
    prior = spline_tilt_prior(nodes)
    costilts, dataset = tilt_test_data(xp)

    norms = list()
    # Use fewer iterations for computational efficiency
    n_test = 10
    for _ in range(n_test):
        parameters = prior.sample()
        prob = model(dataset, **parameters)
        norm = float(trapezoid(trapezoid(prob, costilts), costilts))
        norms.append(norm)

    # Allow 5% error tolerance for numerical integration
    assert xp.max(xp.abs(1 - xp.asarray(norms))) < 0.05


@pytest.mark.parametrize("kind", ["linear", "cubic"])
def test_spline_spin_tilt_interpolation_kind(backend, kind):
    """Test that different interpolation kinds work for SplineSpinTiltIdentical."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinTiltIdentical(nodes=nodes, kind=kind)
    prior = spline_tilt_prior(nodes)
    costilts, dataset = tilt_test_data(xp)

    parameters = prior.sample()
    prob = model(dataset, **parameters)
    norm = float(trapezoid(trapezoid(prob, costilts), costilts))

    # Should be normalized regardless of interpolation kind
    # Allow 5% error tolerance for numerical integration
    assert abs(norm - 1.0) < 0.05


@pytest.mark.parametrize("regularize", [True, False])
def test_spline_spin_tilt_regularization(backend, regularize):
    """Test that regularization option works for SplineSpinTiltIdentical."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinTiltIdentical(nodes=nodes, regularize=regularize)
    prior = spline_tilt_prior(nodes)

    if regularize:
        prior["rmscos_tilt"] = Uniform(0.1, 2.0)

    costilts, dataset = tilt_test_data(xp)
    parameters = prior.sample()

    # Should not raise an error
    prob = model(dataset, **parameters)

    # Check that variable names match expected
    if regularize:
        assert "rmscos_tilt" in model.variable_names
    else:
        assert "rmscos_tilt" not in model.variable_names


def test_spline_spin_tilt_zero_outside_range(backend):
    """Test that SplineSpinTiltIdentical returns zero outside the valid range."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    minimum, maximum = -0.5, 0.5
    model = spin.SplineSpinTiltIdentical(minimum=minimum, maximum=maximum, nodes=nodes)
    prior = spline_tilt_prior(nodes, minimum=minimum, maximum=maximum)

    # Create test data that includes values outside the range
    costilts = xp.linspace(-1, 1, 2000)
    dataset = dict(
        cos_tilt_1=xp.einsum("i,j->ij", costilts, xp.ones_like(costilts)),
        cos_tilt_2=xp.einsum("i,j->ji", costilts, xp.ones_like(costilts)),
    )

    parameters = prior.sample()
    prob = model(dataset, **parameters)

    # Values outside [-0.5, 0.5] should be zero
    prob_np = gwpopulation.utils.to_numpy(prob)
    costilts_np = gwpopulation.utils.to_numpy(costilts)

    assert np.max(prob_np[(costilts_np < minimum) | (costilts_np > maximum)]) == 0.0


def test_spline_spin_tilt_variable_names():
    """Test that variable_names property works for SplineSpinTiltIdentical."""
    nodes = 5
    model = spin.SplineSpinTiltIdentical(nodes=nodes, regularize=False)

    # Should have node positions and values
    expected = [f"cos_tilt{ii}" for ii in range(nodes)] + [
        f"fcos_tilt{ii}" for ii in range(nodes)
    ]
    assert set(model.variable_names) == set(expected)

    # With regularization, should also have rms parameter
    model_reg = spin.SplineSpinTiltIdentical(nodes=nodes, regularize=True)
    expected_reg = expected + ["rmscos_tilt"]
    assert set(model_reg.variable_names) == set(expected_reg)


# Test boundary conditions


def test_spline_magnitude_at_boundaries(backend):
    """Test that SplineSpinMagnitudeIdentical handles boundary values correctly."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinMagnitudeIdentical(minimum=0, maximum=1, nodes=nodes)
    prior = spline_magnitude_prior(nodes)

    # Test with data exactly at boundaries
    dataset = dict(
        a_1=xp.array([[0.0, 0.5, 1.0]]),
        a_2=xp.array([[0.0, 0.5, 1.0]]),
    )

    parameters = prior.sample()
    prob = model(dataset, **parameters)

    # Should get valid probabilities at boundaries
    assert xp.all(prob >= 0)


def test_spline_tilt_at_boundaries(backend):
    """Test that SplineSpinTiltIdentical handles boundary values correctly."""
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    nodes = 5
    model = spin.SplineSpinTiltIdentical(minimum=-1, maximum=1, nodes=nodes)
    prior = spline_tilt_prior(nodes)

    # Test with data exactly at boundaries
    dataset = dict(
        cos_tilt_1=xp.array([[-1.0, 0.0, 1.0]]),
        cos_tilt_2=xp.array([[-1.0, 0.0, 1.0]]),
    )

    parameters = prior.sample()
    prob = model(dataset, **parameters)

    # Should get valid probabilities at boundaries
    assert xp.all(prob >= 0)



import unittest

import numpy as np
import pandas as pd
from bilby.core.prior import PriorDict, Uniform
from bilby.hyper.model import Model

import gwpopulation
from gwpopulation.hyperpe import (
    HyperparameterLikelihood,
    NullHyperparameterLikelihood,
    RateLikelihood,
)
from gwpopulation.models.mass import SinglePeakSmoothedMassDistribution

xp = np


class Likelihoods(unittest.TestCase):
    def setUp(self):
        gwpopulation.set_backend("numpy")
        self.params = dict(a=1, b=1, c=1)
        self.model = lambda dataset, a, b, c: dataset["a"]
        one_data = pd.DataFrame({key: xp.ones(500) for key in self.params})
        self.data = [one_data] * 5
        self.ln_evidences = [0] * 5
        self.selection_function = lambda args: 2.0
        self.conversion_function = lambda args: (args, ["bar"])

    def tearDown(self):
        pass

    def test_hpe_likelihood_requires_posteriors(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(hyper_prior=self.model)

    def test_hpe_likelihood_requires_hyper_prior(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(posteriors=self.data)

    def test_likelihood_pass_sampling_prior_raises_error(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(
                posteriors=self.data,
                hyper_prior=self.model,
                sampling_prior=lambda dataset: 5,
            )

    def test_prior_in_posteriors(self):
        for frame in self.data:
            frame["prior"] = 1
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertTrue(
            xp.array_equal(like.sampling_prior, xp.ones_like(like.data["a"]))
        )

    def test_not_passing_prior(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertEqual(like.sampling_prior, 1)

    def test_hpe_likelihood_set_evidences(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, ln_evidences=self.ln_evidences
        )
        self.assertEqual(like.total_noise_evidence, 0)

    def test_hpe_likelihood_dont_set_evidences(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertTrue(xp.isnan(like.total_noise_evidence))

    def test_hpe_likelihood_set_conversion(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            conversion_function=self.conversion_function,
        )
        self.assertEqual(like.conversion_function("foo"), ("foo", ["bar"]))

    def test_hpe_likelihood_set_selection(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
        )
        self.assertEqual(like.selection_function("foo"), 2.0)

    def test_hpe_likelihood_set_max_samples(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, max_samples=10
        )
        self.assertEqual(like.data["a"].shape, (5, 10))

    def test_hpe_likelihood_log_likelihood_ratio(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertEqual(like.log_likelihood_ratio(self.params), 0.0)

    def test_hpe_likelihood_converts_nan_to_neginf(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=lambda *args, **kwargs: (np.nan, 0),
        )
        self.assertEqual(like.log_likelihood_ratio(self.params), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_noise_likelihood_ratio(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.assertEqual(like.noise_log_likelihood(), 0)

    def test_hpe_likelihood_log_likelihood_equal_ratio_zero_evidence(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.assertEqual(
            like.log_likelihood_ratio(self.params), like.log_likelihood(self.params)
        )

    def test_hpe_likelihood_population_variance_too_large_returns_neginf(self):
        xp.random.seed(10)
        self.data[0]["a"] *= xp.random.uniform(0, 2, self.data[0]["a"].shape)
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, maximum_uncertainty=1e-5
        )
        self.assertEqual(like.log_likelihood_ratio(self.params), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_selection_variance_too_large_returns_neginf(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            maximum_uncertainty=0.1,
            selection_function=lambda *args, **kwargs: (1e-5, 1),
        )
        self.assertEqual(like.log_likelihood_ratio(self.params), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_variance_small_enough_returns_expected(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, maximum_uncertainty=0.1
        )
        like._get_selection_factor = lambda *args, **kwargs: (0, 0)
        self.assertEqual(like.log_likelihood_ratio(self.params), 0.0)

    def test_rate_likelihood_requires_rate(self):
        like = RateLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        with self.assertRaises(KeyError):
            like.log_likelihood_ratio(self.params)

    def test_generate_extra_statistics(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            conversion_function=self.conversion_function,
            ln_evidences=self.ln_evidences,
        )
        self.params["bar"] = None
        new_params = like.generate_extra_statistics(sample=self.params.copy())
        expected = {
            "a": 1,
            "b": 1,
            "c": 1,
            "ln_bf_0": 0.0,
            "ln_bf_1": 0.0,
            "ln_bf_2": 0.0,
            "ln_bf_3": 0.0,
            "ln_bf_4": 0.0,
            "selection": 2.0,
            "selection_variance": 0.0,
            "var_0": 0.0,
            "var_1": 0.0,
            "var_2": 0.0,
            "var_3": 0.0,
            "var_4": 0.0,
            "variance": 0.0,
            "bar": None,
        }
        self.assertDictEqual(expected, new_params)

    def test_generate_rate_posterior_sample_returns_positive(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.assertGreater(like.generate_rate_posterior_sample(self.params), 0)

    def test_resampling_posteriors(self):
        priors = PriorDict(dict(a=Uniform(0, 2), b=Uniform(0, 2), c=Uniform(0, 2)))
        samples = priors.sample(100)
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        new_samples = like.posterior_predictive_resample(samples=samples)
        for key in new_samples:
            self.assertEqual(new_samples[key].shape, like.data[key].shape)

    def test_meta_data(self):
        model = Model([self.model, SinglePeakSmoothedMassDistribution()])
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        expected = dict(
            model=["<lambda>", "SinglePeakSmoothedMassDistribution"],
            data=dict(
                a=np.ones((5, 500)),
                b=np.ones((5, 500)),
                c=np.ones((5, 500)),
            ),
            n_events=5,
            sampling_prior=1,
            samples_per_posterior=500,
        )
        for key in like.meta_data:
            if key == "data":
                for key_2 in like.meta_data[key]:
                    self.assertTrue(
                        np.max(abs(like.meta_data[key][key_2] - expected[key][key_2]))
                        == 0
                    )
            else:
                self.assertEqual(like.meta_data[key], expected[key])


class NullHyperparameterLikelihoodTests(unittest.TestCase):
    def setUp(self):
        gwpopulation.set_backend("numpy")
        self.params = dict(a=1, b=1, c=1)
        self.model = lambda dataset, a, b, c: dataset["a"]
        one_data = pd.DataFrame({key: xp.ones(500) for key in self.params})
        self.data = [one_data] * 5
        self.ln_evidences = [0] * 5
        self.selection_function = lambda args: 2.0

    def tearDown(self):
        pass

    def test_null_likelihood_variance_matches_hyperparameter_likelihood(self):
        """Test that NullHyperparameterLikelihood variance matches HyperparameterLikelihood"""
        # Create both likelihoods with the same parameters
        null_like = NullHyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
        )
        hyper_like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
        )

        # Get the ln_likelihood and variance from both
        null_ln_l, null_variance = null_like.ln_likelihood_and_variance(self.params)
        hyper_ln_l, hyper_variance = hyper_like.ln_likelihood_and_variance(self.params)

        # Check that variances match
        self.assertEqual(null_variance, hyper_variance)

    def test_null_likelihood_returns_zero(self):
        """Test that NullHyperparameterLikelihood returns zero for the likelihood"""
        null_like = NullHyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
        )

        # Get the ln_likelihood
        null_ln_l, null_variance = null_like.ln_likelihood_and_variance(self.params)

        # Check that likelihood is zero
        self.assertEqual(null_ln_l, 0.0)

    def test_null_likelihood_returns_neginf_when_variance_exceeds_threshold(self):
        """Test that NullHyperparameterLikelihood returns -inf when variance > threshold"""
        # Create data with some variation to generate non-zero variance
        xp.random.seed(10)
        varied_data = [self.data[0].copy() for _ in range(5)]
        varied_data[0]["a"] *= xp.random.uniform(0, 2, varied_data[0]["a"].shape)

        # Create likelihood with small maximum_uncertainty threshold
        null_like = NullHyperparameterLikelihood(
            posteriors=varied_data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            maximum_uncertainty=1e-5,
        )

        # Get the log likelihood ratio (which incorporates the variance check)
        log_lr = null_like.log_likelihood_ratio(self.params)

        # Check that it returns -inf
        self.assertEqual(log_lr, np.nan_to_num(-np.inf))



import importlib

import pytest

from gwpopulation import backend as _backend

TEST_BACKENDS = list()
for backend in _backend.SUPPORTED_BACKENDS:
    try:
        importlib.import_module(backend)
    except ImportError:
        continue
    TEST_BACKENDS.append(backend)


@pytest.fixture(params=TEST_BACKENDS)
def backend(request):
    pytest.importorskip(request.param)
    return request.param



import functools
import importlib

import numpy as np
import pytest

from gwpopulation.backend import _np_module, set_backend
from gwpopulation.models.interped import (
    InterpolatedNoBaseModelIdentical,
    _setup_interpolant,
)


@pytest.mark.parametrize("regularize", [True, False])
def test_regularization_works(regularize):
    model = InterpolatedNoBaseModelIdentical(
        parameters=["a_1"], minimum=0, maximum=1, regularize=regularize
    )
    parameters = {key: value for key, value in zip(model.xkeys, np.linspace(0, 1, 10))}
    parameters.update({key: 1.0 for key in model.fkeys})
    if regularize:
        parameters["rmsa"] = 0.1
    values = model.extract_spline_points(parameters)[0]
    if regularize:
        expected = 0.1
    else:
        expected = 1
    assert abs((values**2).mean() ** 0.5 - expected) < 1e-10


def test_regularize_option_adds_rms_variable():
    for regularize in [True, False]:
        model = InterpolatedNoBaseModelIdentical(
            parameters=["a_1"], minimum=0, maximum=1, regularize=regularize
        )
        assert ("rmsa" in model.variable_names) is regularize


def test_setup_interpolant_obeys_backend(backend):
    xp = importlib.import_module(_np_module[backend])
    interpolant = _setup_interpolant(
        nodes=np.linspace(0, 1, 5),
        values=np.linspace(0, 1, 1000),
        backend=xp,
    )
    assert isinstance(interpolant.func.conversion, xp.ndarray)


def test_setup_interpolant_obeys_backend_implicit(backend):
    xp = importlib.import_module(_np_module[backend])
    set_backend(backend)
    interpolant = _setup_interpolant(
        nodes=np.linspace(0, 1, 5),
        values=np.linspace(0, 1, 1000),
    )
    assert isinstance(interpolant.func.conversion, xp.ndarray)



"""
When adding significant new functionality to the package, it is useful to have a place to test and develop.
This subpackage contains experimental new features and models that do not have a finalized API.
Any code in this subpackage is subject to change without warning in subsequent versions.

Current experimental features include:

- cosmology functionality using `wcosmo <https://wcosmo.readthedocs.io>`_.
- :code:`JAX` support for JIT-compiling population models and likelihoods.
- a :code:`numpyro` compatible definition of the :func:`gwpopulation.hyperpe.HyperparameterLikelihood`.
"""



r"""
Searches for gravitational-wave transients are limited by the sensitivity of current detectors.
For this reason it is necessary to quantify the fraction of sources that are expected to be observed
to avoid biases in the population inference.

.. math::

    P_{\rm det}(\Lambda) = \int dd \int d\theta p(d, \theta | \lambda) \Theta(\rho(d) - \rho_{\rm th})

Here :math:`d` is the observed strain data, :math:`\theta` are the parameters of individual sources,
e.g., masses, spins, redshifts, etc., and :math:`\Lambda` are the population parameters.
The quantity :math:`\rho` is a detection statistic, e.g., the signal-to-noise ratio, and :math:`\rho_{\rm th}`
is the threshold for detection.

The most common method to estimate this quantity is to simulate a population of sources from
some reference distribution :math:`p(\theta | \varnothing)` and
compute the fraction of sources that are detected. Using a single reference set of such "injections"
one can estimate :math:`P_{\rm det}(\Lambda)` using Monte Carlo integration.

.. math::

    \hat{P}_{\rm det}(\Lambda) = \frac{1}{N} \sum_{i=1}^N \Theta(\rho_i - \rho_{\rm th})
    \frac{p(\theta_i | \Lambda)}{p(\theta_i | \varnothing)}

Since the detection statistic is independent of the population model, we can remove the :math:`\theta_i`
that don't pass the threshold yielding :math:`M` detected sources.

.. math::

    \hat{P}_{\rm det}(\Lambda) = \frac{1}{N} \sum_{i=1}^M
    \frac{p(\theta_i | \Lambda)}{p(\theta_i | \varnothing)}

This model is implemented in the :class:`gwpopulation.vt.ResamplingVT` class.

A simpler model is to interpolate some expression for

.. math::

    p_{\rm det}(\theta) = \int dd p(d, \theta) \Theta(\rho(d) - \rho_{\rm th})

The quantity :math:`P_{\rm det}(\Lambda)` can be computed by integrating over the
specified :math:`\theta`. This model is implemented in the :class:`gwpopulation.vt.GridVT` class.
Note that the computational cost of this approach scales exponentially with the number of parameters.
"""

import numpy as np
from bilby.hyper.model import Model

from .models.redshift import _Redshift, total_four_volume
from .utils import to_number, trapezoid

xp = np

__all__ = [
    "xp",
    "GridVT",
    "ResamplingVT",
]


class _BaseVT:
    def __init__(self, model, data):
        self.data = {key: xp.asarray(value) for key, value in data.items()}
        if isinstance(model, list):
            model = Model(model)
        elif not isinstance(model, Model):
            model = Model([model])
        self.model = model

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class GridVT(_BaseVT):
    """
    Evaluate the sensitive volume on a grid.

    Parameters
    ----------
    model: callable
        Population model
    data: dict
        The sensitivity labelled :code:`vt` and an entry for each
        parameter to be marginalized over.
    """

    def __init__(self, model, data):
        self.vts = xp.asarray(data.pop("vt"))
        super(GridVT, self).__init__(model=model, data=data)
        self.values = {key: xp.unique(self.data[key]) for key in self.data}
        shape = np.array(list(self.data.values())[0].shape)
        lens = {key: len(self.values[key]) for key in self.data}
        self.axes = {int(np.where(shape == lens[key])[0][0]): key for key in self.data}
        self.ndim = len(self.axes)

    def __call__(self, parameters):
        vt_fac = self.model.prob(self.data, **parameters) * self.vts
        for ii in range(self.ndim):
            vt_fac = trapezoid(
                vt_fac, self.values[self.axes[self.ndim - ii - 1]], axis=-1
            )
        return vt_fac


class ResamplingVT(_BaseVT):
    r"""
    Evaluate the sensitive volume using a set of found injections.

    See `Farr <https://arxiv.org/abs/1904.10879>`_ for details of the formalism.

    There is an option to use the uncertainty-marginalized vt_factor from
    Equation 11 in `Farr <https://arxiv.org/abs/1904.10879>` by setting
    :code:`marginalize_uncertainty = True`, or use the estimator from
    Equation 8 (default behavior).

    We recommend not enabling :code:`marginalize_uncertainty` and setting
    convergence criteria based on uncertainty in total likelihood in
    HyperparameterLikelihood.

    If using :code:`marginalize_uncertainty`: and
    :math:`n_{\rm eff} < 4 n_{\rm events}` we return :code:`np.inf`
    so that the sample is rejected. This condition is also
    enforced if :code:`enforce_convergence=True`.


    Parameters
    ----------
    model: callable
        Population model
    data: dict
        The found injections and relevant meta data
    n_events: int
        The number of events observed
    marginalize_uncertainty: bool (Default: :code:`False`)
        Whether to return the uncertainty-marginalized pdet from Eq 11
        in `Farr <https://arxiv.org/abs/1904.10879>`_. We recommend not to use
        this as it is not completely understood if this uncertainty
        marginalization is correct.
    enforce_convergence: bool (Default: :code:`True`)
        Whether to enforce the condition that :math:`n_{\rm eff} > 4 n_{\rm events}`.
        This flag only acts when marignalize_uncertainty is :code:`False`.
    """

    def __init__(
        self,
        model,
        data,
        n_events=np.inf,
        marginalize_uncertainty=False,
        enforce_convergence=True,
    ):
        super(ResamplingVT, self).__init__(model=model, data=data)
        self.n_events = n_events
        self.total_injections = data.get("total_generated", len(data["prior"]))
        self.analysis_time = data.get("analysis_time", 1)
        self.redshift_model = None
        self.marginalize_uncertainty = marginalize_uncertainty
        self.enforce_convergence = enforce_convergence
        for _model in self.model.models:
            if isinstance(_model, _Redshift):
                self.redshift_model = _model
        if self.redshift_model is None:
            self._surveyed_hypervolume = total_four_volume(
                lamb=0, analysis_time=self.analysis_time
            )

    def __call__(self, parameters):
        r"""
        Compute the expected fraction of detected sources given a
        set of injections for the specified population model.

        Parameters
        ----------
        parameters: dict
            The population parameters

        Returns
        -------
        (mu, vt_factor): float
            The expected number of detections if
            :code:`self.marginalize_uncertainty=False` or the uncertainty-marginalized
            vt_factor if :code:`self.marginalize_uncertainty=True`.
        var: float
            The variance in the estimate of :math:`P_{\rm det}` if
            :code:`self.marginalize_uncertainty=False`.
        """
        if not self.marginalize_uncertainty:
            mu, var = self.detection_efficiency(parameters)
            if self.enforce_convergence:
                _, correction = self.check_convergence(mu, var)
                mu += correction
            return mu, var
        else:
            vt_factor = self.vt_factor(parameters)
            return vt_factor

    def check_convergence(self, mu, var):
        r"""
        Check if the estimate of the detection efficiency has converged
        beyond the threshold of :math:`\frac{\mu^2}{\sigma^2} > 4 n_{\rm events}`.
        """
        converged = mu**2 > 4 * self.n_events * var
        return (
            converged,
            xp.nan_to_num(xp.inf * (1 - converged), nan=0, posinf=xp.inf),
        )

    def vt_factor(self, parameters):
        r"""
        Compute the expected number of detections given a set of injections.

        This is implemented as in `Farr <https://arxiv.org/abs/1904.10879>`_

        .. math::

            \text{vt_factor} = \frac{\mu}{\exp\left(\frac{3 + n_{\rm events}}{2 n_{\rm eff}}\right)}

        Parameters
        ----------
        parameters: dict
            The population parameters

        Returns
        -------
        vt_factor: float
            The uncertainty-marginalized vt_factor
        """
        mu, var = self.detection_efficiency(parameters)
        _, correction = self.check_convergence(mu, var)
        n_effective = mu**2 / var
        vt_factor = mu / xp.exp((3 + self.n_events) / 2 / n_effective)
        vt_factor += correction
        return vt_factor

    def detection_efficiency(self, parameters):
        r"""
        Compute the expected fraction of detections given a set of injections
        and the variance in the Monte Carlo estimate.

        Parameters
        ----------
        parameters: dict
            The population parameters

        Returns
        -------
        mu: float
            The expected fracion of detections :math:`P_{\rm det}`.
        var: float
            The variance in the estimate of :math:`P_{\rm det}`.
        """
        weights = self.model.prob(self.data, **parameters) / self.data["prior"]
        mu = to_number(xp.sum(weights) / self.total_injections, float)
        var = to_number(
            xp.sum(weights**2) / self.total_injections**2
            - mu**2 / self.total_injections,
            float,
        )
        return mu, var

    def surveyed_hypervolume(self, parameters):
        r"""
        The total surveyed 4-volume with units of :math:`{\rm Gpc}^3{\rm yr}`.

        .. math::
            \mathcal{V} = \int dz \frac{dV_c}{dz} \frac{\psi(z)}{1 + z}

        If no redshift model is specified, assume :math:`\psi(z)=1`.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters to compute the volume at

        Returns
        -------
        float
            The volume

        """
        if self.redshift_model is None:
            return self._surveyed_hypervolume
        elif isinstance(self.redshift_model, _Redshift):
            return (
                self.redshift_model.normalisation(parameters) / 1e9 * self.analysis_time
            )



import bilby
import pandas as pd
import pytest
from bilby.hyper.model import Model
from numpyro import handlers

import gwpopulation
from gwpopulation.experimental.numpyro import gwpopulation_likelihood_model


@pytest.mark.parametrize("apply_selection", [True, False])
def test_gwpopulation_likelihood_model_matches_standard(apply_selection):
    """
    Test that gwpopulation_likelihood_model produces the same likelihood
    as the standard HyperparameterLikelihood.log_likelihood_ratio method.
    """
    # Skip if JAX is not available
    pytest.importorskip("jax")
    pytest.importorskip("numpyro")

    # Set backend to JAX for numpyro compatibility
    gwpopulation.set_backend("jax")
    xp = gwpopulation.models.mass.xp
    bilby.core.utils.random.seed(10)
    rng = bilby.core.utils.random.rng

    # Set up the model
    model = Model(
        [
            gwpopulation.models.mass.SinglePeakSmoothedMassDistribution(),
            gwpopulation.models.spin.independent_spin_magnitude_beta,
            gwpopulation.models.spin.independent_spin_orientation_gaussian_isotropic,
            gwpopulation.models.redshift.PowerLawRedshift(),
        ],
        cache=False,
    )
    vt_model = Model(
        [
            gwpopulation.models.mass.SinglePeakSmoothedMassDistribution(),
            gwpopulation.models.spin.independent_spin_magnitude_beta,
            gwpopulation.models.spin.independent_spin_orientation_gaussian_isotropic,
            gwpopulation.models.redshift.PowerLawRedshift(),
        ],
        cache=False,
    )

    # Create synthetic data
    bounds = dict(
        mass_1=(20, 25),
        mass_ratio=(0.9, 1),
        a_1=(0, 1),
        a_2=(0, 1),
        cos_tilt_1=(-1, -1),
        cos_tilt_2=(-1, -1),
        redshift=(0, 2),
        prior=(1, 1),
    )
    posteriors = [
        pd.DataFrame({key: rng.uniform(*bound, 100) for key, bound in bounds.items()})
        for _ in range(10)
    ]
    vt_data = {
        key: xp.asarray(rng.uniform(*bound, 10000)) for key, bound in bounds.items()
    }

    # Create selection function
    selection = gwpopulation.vt.ResamplingVT(vt_model, vt_data, len(posteriors))

    # Create likelihood
    likelihood = gwpopulation.hyperpe.HyperparameterLikelihood(
        hyper_prior=model,
        posteriors=posteriors,
        selection_function=selection if apply_selection else lambda args: 1,
    )

    # Get a prior sample
    priors = bilby.core.prior.PriorDict("priors/bbh_population.prior")
    prior_sample = priors.sample()

    # Get the standard likelihood value
    standard_ln_l = likelihood.log_likelihood_ratio(prior_sample)

    # Get the numpyro likelihood value using a seed handler
    # Note: predictive_resample=False to avoid the stochastic resampling
    with handlers.seed(rng_seed=0):
        numpyro_ln_l = gwpopulation_likelihood_model(
            likelihood,
            prior_sample,
            apply_selection=apply_selection,
            predictive_resample=False,
        )

    # Convert to number if needed
    if hasattr(numpyro_ln_l, "item"):
        numpyro_ln_l = float(numpyro_ln_l.item())
    else:
        numpyro_ln_l = float(numpyro_ln_l)

    # Check that they match
    assert abs(standard_ln_l - numpyro_ln_l) < 1e-6, (
        f"Numpyro likelihood {numpyro_ln_l} does not match "
        f"standard likelihood {standard_ln_l}"
    )
